/*
 * @(#)UIH.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy;

import java.awt.Component;
import java.awt.Container;
import chrriis.uihierarchy.compatibility.ArrayList;
import chrriis.uihierarchy.compatibility.List;

/**
 * The user interface hierarchy core class. This is the base class that allows
 * to create easily some simple to complex user interfaces containment
 * hierarchies.
 * <br><br>
 * The first steps to creating a hierarchy is to get a clear representation of
 * it. Let's take this simple example:
 * <br><br>
 * <code>
 * * Content pane of a root pane container.<br>
 * &nbsp;&nbsp;+ A panel, using a gridBag layout, added to the center.<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;- A label "A label: " in the cell [0, 0], left
 * aligned.<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;- A button "A button" in the cell [0, 1].
 * </code>
 * <br><br>
 * Then to call UIH the same way:
 * <br><br>
 * <code>
 * UIH h = new UIH();<br>
 * h.openNew(getContentPane());<br>
 * &nbsp;&nbsp;h.open().layout(new GridBagLayout()).constrain(BorderLayout.CENTER);<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;h.add("A label: ").constrain("anchor=WEST");<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;h.add(new JButton("A button")).constrain("gridx=0, gridy=1");
 * <br>
 * &nbsp;&nbsp;h.close();<br>
 * h.close();
 * </code>
 * <br><br>
 * When building the hierarchy containment tree, four methods are important:
 * <br>
 * <ul>
 * <li><code>openNew</code>: open a new containment hierarchy.
 * <li><code>open</code>: open a new level in the containment hierarchy.
 * <li><code>add</code>: add a component at the current hierarchy level.
 * <li><code>close</code>: close a level.
 * </ul>
 * The <code>openNew</code>, <code>open</code> and <code>add</code> methods
 * return an object representing the level, on which it is possible to perform
 * additional operations like setting the constraints or the layout.
 * <br>
 * The <code>openNew</code> method assumes that the container it contains
 * already belongs to a parent. Thus, it is used to start the complete
 * hierarchy, or to add a new level for which the container must not be added to
 * the container of the parent level.
 * <br><br>
 * Labels can be created using just a String, and intermediate panels can be
 * omitted. They are traditionaly declared only to be able to add them to the
 * hierarchy or apply layouts, which is not required with <code>UIH</code>.
 * <br><br>
 * Constraints for layouts are passed as an object. If there is a handler
 * registered for the layout and if it can handle this type of constraints, it
 * will be used: some handlers in the package
 * <code>chrriis.uihierarchy.constraints</code> are pre-defined to handle Strings.
 * The example above is using one of them. Note that it is possible to register
 * custom constraints handler for existing or custom layouts. It is also
 * possible to use a handler when invoking the method: the constraints for the
 * component will be obtained from this handler.
 * <br><br>
 * Another possibility to declare the same example is to use the
 * <code>XmlUIHierarchy</code> class (more about this topic is explained in the
 * corresponding class documentation).
 * <br><br>
 * Note that only when the last <code>close()</code> is invoked, the
 * containement hierarchy gets created.
 * <br><br>
 * After the hierarchy is created it is possible to create some accessors for
 * the levels. An accessor allows to manipulate many levels at once, including
 * their sublevels. For example, it is possible to run an action on one or more
 * levels, eventually including all their sublevels. One creates an
 * <code>Accessor</code> by invoking the <code>createAccessor(String)</code>
 * method using some names that were used to <code>map</code> some levels.
 * <br>
 * Apart from running custom actions, accessors already define convenient
 * methods to disable/enable all the components of the levels it contains, or
 * the levels and their sublevels, and to show/hide the components at those
 * levels.
 * <br><br>
 * Generally, a logical set of tasks is to be performed when building a
 * containment hierarchy. It is usually good to follow this sequence:
 * <br><br>
 * <ul>
 * <li> Declare all the references of components needed later on.
 * <li> Create the containment hierarchy using those references and some inline
 * or implicit declarations.
 * <li> Eventually create some accessors to some levels.
 * <li> Add the listeners to some of the components.
 * </ul>
 * More capabilities are provided, like the possibility to <code>print</code>
 * the hierarchy that is under creation. It is a convenient method to facilitate
 * debugging.
 * @version 2.0 2003.11.03
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class UIH {

  /**
   * A reference to this UIH instance. It is mainly used to visually represent
   * the hierarchy tree in the source code, and keep the source format after
   * some IDE automatic source formatting.<br>
   * <br>
   * Here is an example of its use:<br>
   * <code>
   * UIH h = new UIH();<br>
   * h.openNew(getContentPane());<br>
   * h._.open().layout(new GridBagLayout()).constrain(BorderLayout.CENTER);<br>
   * h._._.add("A label: ").constrain("anchor=WEST");<br>
   * h._._.add(new JButton("A button")).constrain("gridx=0, gridy=1");<br>
   * h._.close();<br>
   * h.close();
   * </code>
   */
  public final UIH _ = this;

  /**
   * Construct a hierarchy of components.
   */
  public UIH() {
    this(new UIHConfig());
  }

  /**
   * Construct a hierarchy of components, using the specific configuration.
   * @param uihConfig The configuration to use.
   */
  public UIH(UIHConfig uihConfig) {
    this.uihConfig = uihConfig;
  }

  /**
   * Create an accessor for some mapped levels. If several levels appear more
   * than once in the mapping, they will be considered only once.<br>
   * This method uses the accessor factory currently defined, so one can cast
   * the resulting accessor to a more suitable accessor.
   * @param names The names of the hierarchy levels to get, in a comma separated
   * String if more than one name is to be specified.
   * @return The hierarchy level accessor.
   * @see chrriis.uihierarchy.accessor.AccessorFactory
   */
  public Accessor createAccessor(String names) {
    return getConfig().createAccessor(names);
  }

  /**
   * Create an accessor for some mapped levels, with excluding some. If several
   * levels appear more than once in the mapping, they will be considered only
   * once.<br>
   * This method uses the accessor factory currently defined, so one can cast
   * the resulting accessor to a more suitable accessor.
   * @param names The names of the hierarchy levels to get, in a comma separated
   * String if more than one name is to be specified.
   * @param excludedNames The names of the levels to exclude when creating an
   * accessor on the names. 
   * @return The hierarchy level accessor.
   * @see chrriis.uihierarchy.accessor.AccessorFactory
   */
  public Accessor createAccessor(String names, String excludedNames) {
    return getConfig().createAccessor(names, excludedNames);
  }

/*----------------------------------------------------------------------------*/

  /**
   * Print a level and its sub levels.
   * @param hLevel The hierarchy level to print.
   * @param increment The increment used to display the hierarchy.
   */
  protected void print(HLevel hLevel, int increment) {
    for(int i=0; i<increment; i++) {
      System.out.print("  ");
    }
    System.out.println(hLevel);
    if(hLevel instanceof HParentLevel) {
      HParentLevel parent = (HParentLevel)hLevel;
      List childrenList = parent.getChildrenList();
      for(int i=0; i<childrenList.size(); i++) {
        print((HLevel)childrenList.get(i), increment + 1);
      }
    }
  }

  /**
   * Print the hierarchy to help debugging.
   * @return The hierarchy.
   */
  public UIH print() {
    print(getRootLevel(), 0);
    return this;
  }

  /**
   * Get the container for which the hierarchy is currently being defined, or
   * the last root container that was created if the hierarchy is completed.
   * @return The container of the current hierarchy, or null if there is no
   * hierarchy.
   */
  public Container getContainer() {
    HRootLevel hRootLevel = getRootLevel();
    if(isClosed()) {
      List childrenList = hRootLevel.getChildrenList();
      for(int i=childrenList.size() - 1; i>=0; i--) {
        HLevel hLevel = (HLevel)childrenList.get(i);
        if(hLevel instanceof HParentLevel) {
          return ((HParentLevel)hLevel).getContainer();
        }
      }
      return null;
    }
    return getCurrentParentLevel().getContainer();
  }

  /**
   * A root level of a hierarchy definition.
   * @version 1.0 2003.12.24
   * @author Christopher Deckers (chrriis@brainlex.com)
   */
  protected class HRootLevel extends HParentLevel {
    /**
     * Construct a root level used as root of a hierarchy definition.
     */
    protected HRootLevel() {
      super(UIH.this.getConfig(), null, null);
    }
    /**
     * Set the component by creating one from an object component.
     * @param component The component object.
     */
    protected void setComponent(Object component) {
    }
    /**
     * Realize the level.
     */
    protected void doRealization() {
      List childrenList = getChildrenList();
      for(int i=0; i<childrenList.size(); i++) {
        HLevel child = (HLevel)childrenList.get(i);
        realizeChild(child);
      }
    }
    /**
     * Add a child level to the list of children.
     * @param hLevel The child level to add.
     */
    protected void addChild(HLevel hLevel) {
      setRealized(false);
      super.addChild(hLevel);
    }

    /**
     * Add a child level to the list of children at the specified index.
     * @param index The index at which to add the child.
     * @param hLevel The child level to add.
     */
    protected void addChild(int index, HLevel hLevel) {
      super.addChild(index, hLevel);
    }
    /**
     * Get a String representation of the hierarchy level.
     * @see java.lang.Object#toString()
     */
    public String toString() {
      return "> UIH [Closed: " + isClosed() + "] [Realized: " + isRealized() + "]";
    }
  }

  /** The root level of this hierarchy definition. */
  protected HRootLevel hRootLevel = new HRootLevel();

  /** The current hierarchy parent level. */
  protected HParentLevel currentHParentLevel = hRootLevel;

  /**
   * Indicate whether the hierarchy is closed.
   * @return True if the hierarchy is closed.
   */
  public boolean isClosed() {
    return currentHParentLevel == getRootLevel();
  }

  /**
   * Get the current parent level.
   * @return The parent level that is currently being defined, or the root
   * level.
   */
  protected HParentLevel getCurrentParentLevel() {
    return currentHParentLevel;
  }

  /**
   * Set the current parent level.
   * @param currentHParentLevel The parent level that is currently being
   * defined, or the root level.
   */
  protected void setCurrentParentLevel(HParentLevel currentHParentLevel) {
    this.currentHParentLevel = currentHParentLevel;
  }

  /**
   * Get the root level of this hierarchy definition.
   * @return The root level of this hierarchy definition.
   */
  protected HRootLevel getRootLevel() {
    return hRootLevel;
  }

  /**
   * Return a copy of the root's list of children levels.
   * @return The list of levels.
   */
  protected List getTopLevelsList() {
    return new ArrayList(getRootLevel().getChildrenList());
  }

  /** The last hierarchy level that was added. */
  protected HLevel lastHLevel;

  /**
   * Get the last level that was added, which holds a component or container.
   * @return The last level that was added.
   */
  protected HLevel getLastLevel() {
    return lastHLevel;
  }

  /**
   * Set the last level that was added.
   * @param lastLevel The last added level.
   */
  protected void setLastLevel(HLevel lastLevel) {
    lastHLevel = lastLevel;
  }

  /**
   * Get the last component (or container) that was added.
   * @return The last added component.
   */
  public Component getComponent() {
    HLevel lastHLevel = getLastLevel();
    if(lastHLevel == null) {
      return null;
    }
    return lastHLevel.getComponent();
  }

  /**
   * Start a hierarchy.
   * @param container The container to add to the hierarchy.
   * @return The root node that was created.
   */
  public HRootNode openNew(Object container) {
    HRootNode rootNode = new HRootNode(getConfig(), getCurrentParentLevel(), container);
    setCurrentParentLevel(rootNode);
    setLastLevel(rootNode);
    return rootNode;
  }

  /**
   * Create a node in the hierarchy.
   * @return The node that was created.
   */
  public HNode open() {
    return open(null);
  }

  /**
   * Create a node in the hierarchy with a specific container.
   * @param container The container to add to the hierarchy.
   * @return The node that was created.
   */
  public HNode open(Object container) {
    HNode node = new HNode(getConfig(), getCurrentParentLevel(), container);
    setCurrentParentLevel(node);
    setLastLevel(node);
    return node;
  }

  /**
   * Create a leaf in the hierarchy.
   * @return The leaf that was created.
   */
  public HLeaf add() {
    return add((Object)null);
  }

  /**
   * Create a leaf in the hierarchy with a specific component.
   * @param component The component to add to the hierarchy.
   * @return The leaf that was created.
   */
  public HLeaf add(Object component) {
    UIHConfig uihConfig = getConfig();
    HParentLevel parentLevel = getCurrentParentLevel();
    HLeaf leaf = new HLeaf(uihConfig, parentLevel, component);
    setLastLevel(leaf);
    if(parentLevel == getRootLevel() && uihConfig.isAutoRealizing()) {
      realize();
    }
    return leaf;
  }

  /**
   * A handler of a child level
   * @version 1.0 2003.12.23
   * @author Christopher Deckers (chrriis@brainlex.com)
   */
  public class SingleLevelHandler {
    /** The level to handle. */
    protected HLevel hLevel;
    /**
     * Construct a child level handler.
     * @param hLevel The level to handle.
     */
    protected SingleLevelHandler(HLevel hLevel) {
      this.hLevel = hLevel;
    }
    /**
     * Use the specified constraints.
     * @param constraints The constraints to use.
     */
    public void constrain(Object constraints) {
      hLevel.constrain_(constraints);
    }
  }

  /**
   * Add a child level and its sub levels from an accessor of a single level.
   * This method is usefull when modifying a hierarchy after it was completely
   * constucted, eventually to add levels from another hierarchy.
   * @param accessor The accessor from which to retrieve the hierarchy to add.
   * @return A child level handler, or null if the level that is added is a root
   * node.
   */
  public SingleLevelHandler add(SingleLevelAccessor accessor) {
    HLevel hLevel = accessor.getHLevel();
    HParentLevel parentLevel = getCurrentParentLevel();
    HLevel[] hLevels = accessor.getHLevels();
    for(int i=0; i<hLevels.length; i++) {
      hLevels[i].clearRealizedConstraints();
    }
    parentLevel.addChild(hLevel);
    setLastLevel(hLevel);
    if(parentLevel == getRootLevel() && uihConfig.isAutoRealizing()) {
      realize();
    }
    if(hLevel instanceof HRootNode) {
      return null;
    }
    return new SingleLevelHandler(hLevel);
  }

  /**
   * Close a hierarchy level.
   * @return The hierarchy.
   */
  public void close() {
    if(isClosed()) {
      throw new IllegalStateException("Cannot close a hierarchy that is already at its root level!");
    }
    HParentLevel parent = getCurrentParentLevel().getParent();
    setCurrentParentLevel(parent);
    if(parent == getRootLevel() && getConfig().isAutoRealizing()) {
      realize();
    }
  }

  /**
   * Realize the hierarchy. This happens automatically at the last close
   * invocation unless indicated otherwise in the configuration.
   */
  public void realize() {
    if(!isClosed()) {
      throw new IllegalStateException("Cannot realize a hierarchy that is not closed!");
    }
    HParentLevel rootLevel = getRootLevel();
    if(rootLevel.isRealized()) {
      throw new IllegalStateException("Cannot realize a hierarchy that is already realized!");
    }
    rootLevel.realize();
    getConfig().processLevelsDebugging();
  }

/*----------------------------------------------------------------------------*/

  /** The configuration of this instance. */
  protected UIHConfig uihConfig;

  /**
   * Get the configuration of this hierarchy instance. For defaults each
   * instance share, they can be changed by using the static methods of the
   * <code>UIHClassConfig</code> class.
   * @return The configuration of this hierarchy.
   */
  public UIHConfig getConfig() {
    return uihConfig;
  }

/*----------------------------------------------------------------------------*/

  /**
   * Add some levels to the debugging process, using default parameters.
   * @param names The names of the mapped levels to debug.
   */
  public void debug(String names) {
    debug(names, null);
  }

  /**
   * Add some levels to the debugging process, using default parameters.
   * @param names The names of the mapped levels to debug.
   * @param parameters The parameters to use to debug, which can actually be a
   * debugger to use.
   */
  public void debug(String names, Object parameters) {
    getConfig().addLevelsDebugging(names, parameters);
  }

}
